/*
 *  Copyright (c) 1992, 1993 John E. Davis  (davis@amy.tch.harvard.edu)
 *  All Rights Reserved.
 */
#include <stdio.h>
#include <string.h>

#ifndef sequent
# include <stdlib.h>
#endif

#include "slang.h"
#include "malloc.h"

/* We need a define a rule for upperand lower case chars that user cannot
   change!  This could be a problem for international chars! */

#define UPPER_CASE_KEY(x) (((x) >= 'a') && ((x) <= 'z') ? (x) - 32 : (x))
#define LOWER_CASE_KEY(x) (((x) >= 'A') && ((x) <= 'Z') ? (x) + 32 : (x))



char Key_Buffer[13];
char *Key_Bufferp = Key_Buffer;
int Key_TimeOut_Flag = 0;	       /* true if more than 1 sec has elapsed
                                          without key in multikey sequence */


/* special self insert key type */
Key_Type Self_Insert_KeyType;

#define MAX_KEYMAP_NAME_LEN 8

typedef struct KeyMap_List_Type
{
   char name[MAX_KEYMAP_NAME_LEN + 1];
   Key_Type *keymap;
}
KeyMap_List_Type;

/* This is arbitrary but I have got to start somewhere */
#ifdef msdos
#define MAX_KEYMAPS 10
#else
#define MAX_KEYMAPS 30
#endif

KeyMap_List_Type KeyMap_List[MAX_KEYMAPS];   /* these better be inited to 0! */

static Key_Type *malloc_key(unsigned char *str)
{
   Key_Type *neew;

   if (NULL == (neew = (Key_Type *) MALLOC(sizeof(Key_Type))))
     {
	SLang_doerror("malloc_key: malloc error.");
	return (NULL);
     }
   MEMCPY((char *) neew->str, (char *) str, (int) *str);
   return(neew);
}

void add_keymap(char *name, Key_Type *map)
{
   int i;

   for (i = 0; i < MAX_KEYMAPS; i++)
     {
	if (KeyMap_List[i].keymap == NULL)
	  {
	     KeyMap_List[i].keymap = map;
	     MEMCPY((char *) KeyMap_List[i].name, (char *) name, MAX_KEYMAP_NAME_LEN);
	     KeyMap_List[i].name[MAX_KEYMAP_NAME_LEN] = 0; 
	     return;
	  }
     }
   SLang_doerror ("Keymap quota exceeded");
}

static VOID *find_function(char *name)
{
   SLKeymap_Fun_Type *fp;

   fp = SLkeymap_functions;

   while((fp != NULL) && (fp->name != NULL))
     {
	if ((*fp->name == *name) && !strcmp(fp->name, name)) return((VOID *) fp->f);
	fp++;
     }
    return(NULL);
}

/* convert things like "^A" to 1 etc... The 0th char is the strlen */
static char *process_keystring(char *s)
{
   static char str[30];
   unsigned char ch;
   int i;

    i = 1;
    while (*s != 0)
      {
          ch = (unsigned char) *s++;
          if (*s) ch = UPPER_CASE_KEY(ch);
          if (ch == '^')
            {
                ch = *s++;
                if (ch == 0)
                  {
                      str[i++] = '^';
                      break;
                  }
	       if (ch == '?') ch = 127; else ch = ch - 'A' + 1;
            }

          str[i++] = ch;
      }
    str[0] = i;
    return(str);
}


static char *Define_Key_Error = "Inconsistency in define key.";

static void define_key1(char *s, VOID *f, char type, Key_Type *keymap)
{
   int cmp, i, m, n, len;
   Key_Type *key, *last, *neew;
   unsigned char *str;

   str = (unsigned char *) process_keystring(s);
   if (1 == (n = str[0])) return;

   i = *(str + 1);
   key = (Key_Type *) &(keymap[i]);

   if (n == 2)
     {
	if (key->next != NULL)
	  {
	     SLang_doerror(Define_Key_Error);
	     return;
	  }
	
	/* copy keymap uses the pointers so do not free these.
	if (key->type == F_INTERPRET) FREE(key->f); */
	
	key->f = f;
	key->type = type;
	MEMCPY((char *) key->str, (char *) str, 2);
	return;
     }

   /* insert the key definition */
   while(1)
     {
	last = key;
	key = key->next;
	if ((key == NULL) || (key->str == NULL)) cmp = -1;
	else
	  {
	     m = key->str[0];
	     len = m;
	     if (m > n) len = n;
	     cmp = MEMCMP((char *) (str + 1), (char *)(key->str + 1), len - 1);
	     if ((m != n) && !cmp)
	       {
		  SLang_doerror(Define_Key_Error);
		  return;
	       }
	     if ((m == n) && (!cmp)) cmp = MEMCMP((char *) str, (char *) key->str, n);
	  }

	if (cmp == 0)
            {
	       /* I do not dare do this since copy keymap keeps the pointers 
	       if (key->type == F_INTERPRET) FREE(key->f); */
	       key->f = f;
	       key->type = type;
	       return;
            }
          else if (cmp < 0)
            {
	       neew = malloc_key(str);

	       neew -> next = key;
	       last -> next = neew;

	       neew->f = f;
	       neew->type = type;
	       return;
            } /* else cmp > 0 */
      }
}

void define_key(char *s, char *funct, Key_Type *keymap)
{
   VOID *f;
   char type;

   f = find_function(funct);
   if (f == NULL)                      /* assume interpreted */
     {
	if (NULL == (f = (VOID *) MALLOC(strlen(funct) + 1)))
	  {
	     exit_error("Malloc error in define key.");
	  }

	strcpy((char *) f, funct);
	type = F_INTERPRET;
     }
   else type = F_INTRINSIC;

   define_key1(s, f, type, keymap);
}


int key_interpret(Key_Type *key)
{
   char *rtc_save;
   char *str;
   int ret, ch;
   
   if (key->type == F_INTRINSIC)
     {
	ret = (int) ((int (*)()) key->f)();
     }
   else
     {
	str = (char *) key->f;
	
	if (*str == ' ')
	  {
	     /* string to insert */
	     insert_string(str + 1);
	  }
	else if ((*str == '@') && (0 != *(str + 1)))
	  {
	     rtc_save = Read_This_Character;
	     
	     Read_This_Character = str + 1;
	     do
	       {
		  ch = jed_getkey();
		  do_key((char) ch);
	       }
	     while (Read_This_Character != NULL);
	     Read_This_Character = rtc_save;
	  }
	
	else if (!SLang_execute_function(str)) SLang_load_string(str);
	ret = 1;
     }
   
   Last_Key_Function = key->f;
   return(ret);
}

#define XKEY(key)  key_interpret((key))

static Key_Type *do_key2(char ch)
{
   register Key_Type *key, *next;
   int i;
   unsigned char ch1; 
   register unsigned char chi, chup, chlow;

   Last_Command_Char = (int) ch;

   i = ch1 = (unsigned char) ch;

   /* if metachar is -1 and char has hi bit set then it self inserts. */
   if ((Meta_Char == -1) && (i > 127))
     {
	return(&Self_Insert_KeyType);
     }

   i = i & 0x7F;

   key = (Key_Type *) &((CBuf->keymap)[i]);

   /* if the next one is null, then we know this is it. */
   if (key->next == NULL) 
     {
	/* check its uppercase counterpart if null and return it */
	if ((key->f == NULL) && (i >= 'a') && (i <= 'z')) key = (Key_Type *) &((CBuf->keymap)[i - 32]);
	return(key);
     }
   
   key = key->next;
   Key_TimeOut_Flag = 1;
   Last_Command_Char = jed_getkey();
   
   i = 2;			       /* 0 is keylen, 1 matches, so 2 */
   ch1 = (unsigned char) Last_Command_Char;
   chup = UPPER_CASE_KEY(ch1); chlow = LOWER_CASE_KEY(ch1);

   while(key != NULL)
     {
	if (KeyBoard_Quit) break;
	chi = key->str[i];
	if ((chup == chi) || (chlow == chi))
	  {
	     if (*key->str == i + 1)
	       {
		  /* look for exact match if possible */
		  if (ch1 != chi)
		    {
		       next = key->next;
		       while (next != NULL)
			 {
			    if (next->str[i] == ch1)
			      {
				 if (next->str[0] == i + 1) return(next);
				 break;
			      }
			    next = next->next;
			 }
		    }
		  return(key);
	       }

	     /* before reading a new key, check to see if it is lowercase
	        and try to find a match for it.  Note that this works because
	        only the LAST character in a key sequence can be lowercase */

	     else if ((ch1 == chlow) && (ch1 != chup))
	       {
		  next = key->next;
		  /* look for the lowercase one */
		  while (next != NULL)
		    {
                        if ((next->str[i] == chlow) &&
                            (next->str[0] == i + 1)) return(next);
		       next = next->next;
		    }
	       }

	     Last_Command_Char = ch1 = jed_getkey();
	     i++;
	     chup = UPPER_CASE_KEY(ch1);
	     chlow = LOWER_CASE_KEY(ch1);
	     continue;
	  }
	else if (chlow < chi) break;  /* not found */

	/* else */
	key = key->next;
     }
   /* not found */
   return(NULL);
}

static int do_key1(char ch)
{
   Key_Type *key;
   int repeat;

   Key_TimeOut_Flag = 0;
   key = do_key2(ch);
   if ((key != NULL) && (key->f != NULL))
     {

	if (Repeat_Factor == NULL) return XKEY(key);
	repeat = *Repeat_Factor;
	Suspend_Screen_Update = 1;

	/* some routines may use the repeat factor as a prefix argument */
	while (repeat-- > 0)
	  {
	     if (KeyBoard_Quit || (*Error_Buffer) ||
		 (Repeat_Factor == NULL)) break;
	 /* (!Executing_Keyboard_Macro && (Repeat_Factor == NULL))) break; */
	     XKEY(key);
	  }
	Repeat_Factor = NULL;

	return(1);
     }
   else if (!Executing_Keyboard_Macro && !KeyBoard_Quit)
     {
	beep();
	flush_input();
     }
   if (KeyBoard_Quit) kbd_quit();
   return(0);
}

int do_key(char ch)
{
   int ret;
   ret = do_key1(ch);
   
   /* keys get buffered up in jed_getkey.  This just resets the pointer. */
   if (!Executing_Keyboard_Macro)
     {
	Key_Bufferp = Key_Buffer;
	*Key_Bufferp = 0;
     }
   return ret;
}



void do_jed(void)
{
   char key;
   char *name;
   Buffer *tthis = CBuf;
   
   Key_TimeOut_Flag = 0;
   /* keys get buffered up in jed_getkey.  This just resets the pointer.
      This one is here because we might be calling this routine from the 
      minibuffer and might not have hit the same block in do_key(). 
      */
   if (!Executing_Keyboard_Macro)
     {
	Key_Bufferp = Key_Buffer;
	*Key_Bufferp = 0;
     }
   key = jed_getkey();
   Last_Command_Char = (int) key;

   /* Mark Undo boundary now because tthis might not be valid later */
   mark_undo_boundary(tthis);
   Repeat_Factor = NULL;

   if (do_key(key)) JWindow->trashed = 1;

    /* internal editing commands may have selected a different buffer
       so put it back. */
    if (CBuf != JWindow->buffer)
      {
	 if (buffer_exists(JWindow->buffer)) name = JWindow->buffer->name; else name = "*scratch*";
	 switch_to_buffer_cmd(name);
      }
}

void jed(void)
{

   Self_Insert_KeyType.f = (VOID *) ins_char_cmd;
   Self_Insert_KeyType.type = F_INTRINSIC;
   Self_Insert_KeyType.next = NULL;

   /* This routine is called from main.  So before actually strting, check
      one more hook to just to make sure  all things are go. */
   SLang_run_hooks("jed_startup_hook", NULL, NULL);

   Jump_Buffer_Ptr = &Jump_Buffer;
   
   if (setjmp(Jump_Buffer.b) != 0)
     {
	SLang_restart(1);   /* just in case */
     }

   if (CBuf != JWindow->buffer)
     {
	switch_to_buffer(JWindow->buffer);
	window_buffer(CBuf);
     }
   JWindow->trashed = 1;
   update((Line *) NULL, 0, 0);
   Read_This_Character = NULL;
   while(1)
     {
	Suspend_Screen_Update = 0;
	do_jed();
	if (!KeyBoard_Quit && (CBuf->flags & BUFFER_TRASHED)
	    && (!Cursor_Motion)) CBuf->hits += 1;
	if (CBuf->hits > User_Vars.max_hits)
	  {
	     auto_save_buffer(CBuf);
	     check_buffers();   /* check files on disk to see if they are recent */
	  }
	
	if ((Last_Key_Function != (VOID *) ins_char_cmd)
	    || (JWindow->trashed) || (JWindow != JWindow->next) ||
	    (Mini_Ghost) || (*Message_Buffer) || (*Error_Buffer))
	  {
	     update((Line *) NULL, 0, 0);
	  }
     }
}

int digit_arg(void)
{
    static int repeat;
    char buf[20], key;
    int i;

    i = 0;
    buf[i++] = (char) Last_Command_Char;
    while(1)
      {
	 buf[i] = 0;
	 /* message(buf); */
	 /*
	 if (!input_pending(&Number_Zero) && !Executing_Keyboard_Macro)
	   {
	      flush_message(buf);
	   }
	      */
	 Key_TimeOut_Flag = 1;
	 key = jed_getkey();
	 if ((key < '0') || (key > '9')) break;
	 buf[i++] = key;
      }
   /*
   if (!input_pending(&Number_Zero) && !Executing_Keyboard_Macro)
     {
	flush_message(buf);
     } */
   repeat = atoi(buf);
   Repeat_Factor = &repeat;
   do_key(key);
   return(1);
}

void undefine_key(char *s, Key_Type *keymap)
{
   int n, i;
   Key_Type *key, *next, *last, *key_root;
   unsigned char *str;

   str = (unsigned char *) process_keystring(s);

   if (0 == (n = *str++ - 1)) return;
   i = *str;
   key = key_root = (Key_Type *) &(keymap[i]);

   last = key_root;
   key = key_root->next;
   while (key != NULL)
     {
	next = key->next;
	if (!MEMCMP((char *)(key->str + 1), (char *) str, n))
	  {
	     /* I do not do this since copy_keymap does not malloc new ones 
	     if (key->type == F_INTERPRET) FREE(key->f); */
	     FREE(key);
	     last->next = next;
	  }
	else last = key;
	key = next;
     }

   if (n == 1)
     {
	*key_root->str = 0;
	key_root->f = NULL;
	key_root->type = 0;
     }
}

char *make_keystring(unsigned char *s)
{
   static char buf[80];
   char *b;
   int n;

   b = buf;
   n = *s++ - 1;
   while (n--)
     {
	if (*s < 32)
	  {
	     *b++ = '^';
	     *b++ = *s + 'A' - 1;
	  }
	else *b++ = *s;
	s++;
     }
   *b = 0;
   return(buf);
}

int which_key(char *f)
{
   int num = 0, i;
   Key_Type *key, *key_root;
   VOID *fp;
   char type;
   unsigned char buf[5];

   if (NULL == (fp = (VOID *) find_function(f))) type = F_INTERPRET;
   else type = F_INTRINSIC;

   i = 128;
   key_root = (Key_Type *) CBuf->keymap;
   while (i--)
     {
	key = key_root->next;
	if ((key == NULL) && (type == key_root->type) &&
	    (((type == F_INTERPRET) && (!strcmp((char *) f, (char *) key_root->f)))
	    || ((type == F_INTRINSIC) && (fp == key_root->f))))
	  {
	     *buf = 2;
	     *(buf + 1) = 128 - 1 - i;
	     SLang_push_string(make_keystring(buf));
	     num++;
	  }
	while (key != NULL)
	  {
	     if ((key->type == type) &&
		 (((type == F_INTERPRET) && (!strcmp((char *) f, (char *) key->f)))
		 || ((type == F_INTRINSIC) && (fp == key->f))))
	       {
		  SLang_push_string(make_keystring(key->str));
		  num++;
	       }
	     key = key->next;
	  }
	key_root++;
     }
   return(num);
}


static char *find_function_string (VOID *f)
{
   SLKeymap_Fun_Type *fp;

   fp = SLkeymap_functions;

   if (f == (VOID *) ins_char_cmd) return "self_insert_cmd";
   
   while((fp != NULL) && (fp->name != NULL))
     {
	if ((VOID *) fp->f == f) return fp->name;
	fp++;
     }
   return NULL;
}


static void dump_this_binding (unsigned char *s, VOID *f, int type)
{
   unsigned char ch;
   char *str,  ctrl[2];
   char *fun;
   int n, len;
   
   ctrl[0] = '^';
   len = *s++ - 1;;
   
   while (len-- > 0)
     {
	n = 1;
        ch = *s++;
	if (ch == 127) 
	  {
	     str = "DEL"; n = 3;
	  }
	else if (ch > ' ') str = (char *) &ch;
	else if (ch == 27) 
	  {
	     str = "ESC";
	     n = 3;
	  }
	else if (ch == ' ')
	  {
	     str = "SPACE";
	     n = 5;
	  }
	else if (ch == '\t')
	  {
	     str = "TAB"; n = 3;
	  }
	else
	  {
	     str = ctrl;
	     *(str + 1) = ch + '@';
	     n = 2;
	  }
	quick_insert((unsigned char *) str, n);
	ins(' ');
     }
   ins_char_n_times('\t', 3);
   
   if (type == F_INTRINSIC)
     {
   	fun = find_function_string (f);
     }
   else fun = (char *) f;

   insert_string (fun);
   newline();
}

void  dump_bindings(char *map)
{
   int i;
   Key_Type *next, *key_root;

   if (NULL == (key_root = find_keymap(map)))
     {
	SLang_doerror("Keymap undefined.");
	return;
     }
   
   for (i = 0; i < 128; i++)
     {
	next = key_root->next;
	if (next != NULL)
	  {
	     while (next != NULL)
	       {
		  dump_this_binding (next->str, next->f, next->type);
		  next = next->next;
	       }
	  }
	else if (key_root->f != NULL) dump_this_binding (key_root->str, key_root->f, key_root->type);
	key_root++;
     }
}


Key_Type *copy_keymap(Key_Type *km)
{
   int i;
   Key_Type *neew, *old, *new_root;

   if (NULL == (new_root = (Key_Type *) CALLOC(128, sizeof(Key_Type))))
     {
	exit_error("Malloc error in copy keymap.");
     }

   for (i = 0; i < 128; i++)
     {
	old = &(km[i]);
	neew = &(new_root[i]);
	neew->f = old->f;
	neew->type = old->type;
	MEMCPY((char *) neew->str, (char *) old->str, (int) *old->str);

	old = old->next;
	while (old != NULL)
	  {
	     neew->next = malloc_key((unsigned char *) old->str);
	     neew = neew->next;
	     neew->f = old->f;
	     neew->type = old->type;
	     old = old->next;
	  }
	neew->next = NULL;
     }
   return(new_root);
}


char *find_key(int *ret)
{
   char *fstr = NULL;
   Key_Type *key;

   *ret = 0;
   Key_TimeOut_Flag = 0;
   Key_Bufferp = Key_Buffer;
   if (NULL == (key = do_key2(jed_getkey()))) return(NULL);
   if (key->type == F_INTRINSIC)
     {
	*ret = 1;
	fstr = find_function_string (key->f);
     }
   else fstr = (char *) key->f;

   return(fstr);
}

Key_Type *create_keymap(char *name)
{
   Key_Type *neew;
   if (NULL != (neew = copy_keymap(Global_Map))) add_keymap(name, neew);
   return neew;
}

Key_Type *find_keymap(char *name)
{
   int i;

   for (i = 0; i < MAX_KEYMAPS; i++)
     {
	if (!strncmp(KeyMap_List[i].name, name, 8)) return KeyMap_List[i].keymap;
     }
   return(NULL);
}

void use_keymap(char *name)
{
   Key_Type *map;
   if (NULL == (map = find_keymap(name)))
     {
	SLang_doerror("Unknown keymap.");
     }
   else (CBuf->keymap = map);
}

char *what_keymap()
{
   int i;

   for (i = 0; i < MAX_KEYMAPS; i++)
     {
	if (CBuf->keymap == KeyMap_List[i].keymap) return KeyMap_List[i].name;
     }
   return "";
}

void set_abort_char(int *c)
{
   char str[2];
   Key_Type *map;
   int i;
#ifdef pc_system
   char ch, *s, *scan = "@#$%^&*()_+?IQWERTYUIOP[]!!ASDFGHJKL!!!!\\ZXCVBNM";
   
   ch = 64 + (char) *c;
   s = scan; while (*s && (*s != ch)) s++;
   if (*s == 0) return;
   Abort_Char = (int) (s - scan) + 0x03;
#else
   Abort_Char = *c;
#endif
   reset_tty();
   init_tty();
   str[0] = *c; str[1] = 0;
    
   for (i = 0; i < MAX_KEYMAPS; i++)
     {
	if ((map = KeyMap_List[i].keymap) == NULL) continue;
	undefine_key(str, map);
	define_key1(str, (VOID *) kbd_quit, F_INTRINSIC, map);
     }
}


static SLKeymap_Fun_Type *Flist_Context;
static int Flist_Context_Len;

#define MAX_USER_FLIST 100
static char *Slang_Functions[MAX_USER_FLIST];
static char **Slang_Flist_Context;


int next_function_list(char *buf)
{
   SLKeymap_Fun_Type *tthis;
   char *name, **max;
   
   while (1)
     {
	tthis = Flist_Context;
	name = tthis->name;
	if (name == NULL) break;
	Flist_Context++;
	if (!Flist_Context_Len || !strncmp(buf, name, Flist_Context_Len))
	  {
	     strcpy(buf, name);
	     return(1);
	  }
     }
   
   max = Slang_Functions + MAX_USER_FLIST;
   while (Slang_Flist_Context < max)
     {
	name = *Slang_Flist_Context;
	if (name == NULL) return(0);
	name++;  /* skip past hash mark */
	Slang_Flist_Context++;
	if (!Flist_Context_Len || !strncmp(buf, name, Flist_Context_Len))
	  {
	     strcpy(buf, name);
	     return(1);
	  }
     }
   return(0);
}

int open_function_list(char *buf)
{
   Flist_Context = SLkeymap_functions;
   Slang_Flist_Context = Slang_Functions;
   Flist_Context_Len = strlen(buf);
   return next_function_list(buf);
}

void add_to_completion(char *name)
{
   char **p, *n;
   static char **last = Slang_Functions;
   
   n = SLang_find_name(name);
   if (n == NULL)
     {
	SLang_doerror("Undefined Symbol.");
	return;
     }
   
   p = last;  /* Slang_Functions; */
   
   while (p < Slang_Functions + MAX_USER_FLIST)
     {
	if (*p == NULL) 
	  {
	     *p = n;
	     last = ++p;
	     return;
	  }
	p++;
     }
   SLang_doerror("Completion Quota Exceeded.");
}

int is_internal(char *f)
{
   if (NULL == find_function(f)) return(0);
   return 1;
}
